from collections import OrderedDict
from datetime import date
from datetime import datetime
from datetime import timedelta

import pytest
from hypothesis import given
from hypothesis import strategies as st
from more_itertools import pairwise

from sdlon.date_utils import _get_employment_from_date
from sdlon.date_utils import date_to_datetime
from sdlon.date_utils import format_date
from sdlon.date_utils import datetime_to_sd_date
from sdlon.date_utils import gen_cut_dates
from sdlon.date_utils import gen_date_intervals
from sdlon.date_utils import get_employment_datetimes
from sdlon.date_utils import is_midnight
from sdlon.date_utils import sd_to_mo_termination_date
from sdlon.date_utils import to_midnight


@given(st.dates())
def test_date_to_datetime(d: date) -> None:
    dt = date_to_datetime(d)
    assert isinstance(dt, datetime)
    assert d.year == dt.year
    assert d.month == dt.month
    assert d.day == dt.day
    assert 0 == dt.hour
    assert 0 == dt.minute
    assert 0 == dt.second
    assert 0 == dt.microsecond


@st.composite
def from_to_datetime(draw):
    """Generate date-intervals from 1930-->2060, where from < to."""
    from_datetimes = st.datetimes(
        min_value=datetime(1930, 1, 1),
        max_value=datetime(2050, 1, 1),
    )
    min_datetime = draw(from_datetimes)

    to_datetimes = st.datetimes(
        min_value=min_datetime + timedelta(seconds=1),
        max_value=datetime(2060, 1, 1),
    )
    max_datetime = draw(to_datetimes)

    return min_datetime, max_datetime


class TestSdToMoTerminationDate:
    def test_assert_string(self):
        with pytest.raises(AssertionError):
            sd_to_mo_termination_date(list())

    def test_assert_date_format_string(self):
        with pytest.raises(AssertionError):
            sd_to_mo_termination_date("invalid string")

    def test_assert_invalid_date(self):
        with pytest.raises(AssertionError):
            sd_to_mo_termination_date("2021-13-01")
        with pytest.raises(AssertionError):
            sd_to_mo_termination_date("2021-12-32")


@pytest.mark.parametrize(
    "emp_date,emp_dep_date,emp_status_date,prof_date,working_time_date,expected_date",
    [
        (
            datetime(2020, 1, 1),
            datetime(2001, 1, 1),
            datetime(2002, 1, 1),
            datetime(2003, 1, 1),
            datetime(2004, 1, 1),
            datetime(2020, 1, 1),
        ),
        (
            datetime(2001, 1, 1),
            datetime(2021, 1, 1),
            datetime(2002, 1, 1),
            datetime(2003, 1, 1),
            datetime(2004, 1, 1),
            datetime(2021, 1, 1),
        ),
        (
            datetime(2001, 1, 1),
            datetime(2002, 1, 1),
            datetime(2022, 1, 1),
            datetime(2003, 1, 1),
            datetime(2004, 1, 1),
            datetime(2022, 1, 1),
        ),
        (
            datetime(2001, 1, 1),
            datetime(2002, 1, 1),
            datetime(2003, 1, 1),
            datetime(2023, 1, 1),
            datetime(2004, 1, 1),
            datetime(2023, 1, 1),
        ),
        (
            datetime(2020, 1, 1),
            datetime(2002, 1, 1),
            datetime(2003, 1, 1),
            datetime(2004, 1, 1),
            datetime(2024, 1, 1),
            datetime(2024, 1, 1),
        ),
    ],
)
def test_get_from_date_return_max_date(
    emp_date: datetime,
    emp_dep_date: datetime,
    emp_status_date: datetime,
    prof_date: datetime,
    working_time_date: datetime,
    expected_date: datetime,
):
    employment = OrderedDict(
        [
            ("EmploymentDate", format_date(emp_date)),
            (
                "EmploymentDepartment",
                OrderedDict([("ActivationDate", format_date(emp_dep_date))]),
            ),
            (
                "EmploymentStatus",
                OrderedDict([("ActivationDate", format_date(emp_status_date))]),
            ),
            ("Profession", OrderedDict([("ActivationDate", format_date(prof_date))])),
            (
                "WorkingTime",
                OrderedDict([("ActivationDate", format_date(working_time_date))]),
            ),
        ]
    )

    from_date = _get_employment_from_date(employment)

    assert from_date == expected_date


def test_get_from_date_always_return_date():
    assert _get_employment_from_date(OrderedDict()) == datetime.min


@pytest.mark.parametrize(
    "emp_date,act_date,exp_datetime",
    [
        ("1960-01-01", "1970-01-01", datetime(1970, 1, 1)),
        ("1970-01-01", "1960-01-01", datetime(1970, 1, 1)),
        ("1970-01-01", "1970-01-01", datetime(1970, 1, 1)),
    ],
)
def test_get_employment_from_date_when_status_is_leave(
    emp_date,
    act_date,
    exp_datetime,
):
    employment = OrderedDict(
        [
            (
                "EmploymentDate",
                emp_date,
            ),
            (
                "AnniversaryDate",
                "2004-08-15",
            ),
            (
                "EmploymentStatus",
                OrderedDict(
                    [
                        (
                            "EmploymentStatusCode",
                            "3",
                        ),
                        (
                            "ActivationDate",
                            act_date,
                        ),
                        (
                            "DeactivationDate",
                            "9999-12-31",
                        ),
                    ]
                ),
            ),
        ]
    )

    datetime_from, datetime_to = get_employment_datetimes(employment)

    assert datetime_from == exp_datetime


@pytest.mark.parametrize(
    "deactivation_date,exp_datetime",
    [
        ("1960-01-01", datetime(1960, 1, 1)),
        ("1970-01-01", datetime(1970, 1, 1)),
    ],
)
def test_get_employment_to_date_when_status_is_leave(
    deactivation_date,
    exp_datetime,
):
    employment = OrderedDict(
        [
            (
                "EmploymentDate",
                "1970-01-01",
            ),
            (
                "AnniversaryDate",
                "2004-08-15",
            ),
            (
                "EmploymentStatus",
                OrderedDict(
                    [
                        (
                            "EmploymentStatusCode",
                            "3",
                        ),
                        (
                            "ActivationDate",
                            "1975-01-01",
                        ),
                        (
                            "DeactivationDate",
                            deactivation_date,
                        ),
                    ]
                ),
            ),
        ]
    )

    datetime_from, datetime_to = get_employment_datetimes(employment)

    assert datetime_to == exp_datetime


@pytest.mark.parametrize(
    "datetime,expected",
    [
        [datetime(1960, 1, 1, 0, 0, 0, 0), datetime(1960, 1, 1, 0, 0, 0, 0)],
        [datetime(1960, 1, 1, 0, 0, 0, 1), datetime(1960, 1, 1, 0, 0, 0, 0)],
        [datetime(1960, 1, 1, 8, 0, 0, 0), datetime(1960, 1, 1, 0, 0, 0, 0)],
        [datetime(1960, 1, 1, 23, 59, 59, 999), datetime(1960, 1, 1, 0, 0, 0, 0)],
        [datetime(1960, 1, 2, 0, 0, 0, 0), datetime(1960, 1, 2, 0, 0, 0, 0)],
    ],
)
def test_to_midnight_parameterized(datetime, expected):
    assert to_midnight(datetime) == expected


@given(datetime=st.datetimes())
def test_to_midnight(datetime):
    midnight = to_midnight(datetime)
    assert midnight.date() == datetime.date()
    assert midnight.hour == 0
    assert midnight.minute == 0
    assert midnight.second == 0
    assert midnight.microsecond == 0


@pytest.mark.parametrize(
    "datetime,expected",
    [
        [datetime(1960, 1, 1, 0, 0, 0, 0), True],
        [datetime(1960, 1, 1, 0, 0, 0, 1), False],
        [datetime(1960, 1, 1, 8, 0, 0, 0), False],
        [datetime(1960, 1, 1, 23, 59, 59, 999), False],
        [datetime(1960, 1, 2, 0, 0, 0, 0), True],
    ],
)
def test_is_midnight(datetime, expected):
    assert is_midnight(datetime) is expected


@given(datetime=st.datetimes())
def test_to_midnight_is_midnight(datetime):
    assert is_midnight(to_midnight(datetime))


@pytest.mark.parametrize(
    "from_date,to_date,expected",
    [
        (
            datetime(1960, 1, 1, 8, 0, 0),
            datetime(1960, 1, 1, 9, 0, 0),
            [(datetime(1960, 1, 1, 8, 0, 0), datetime(1960, 1, 1, 9, 0, 0))],
        ),
        (
            datetime(1960, 1, 1, 8, 0, 0),
            datetime(1960, 1, 2, 9, 0, 0),
            [
                (datetime(1960, 1, 1, 8, 0, 0), datetime(1960, 1, 2, 0, 0, 0)),
                (datetime(1960, 1, 2, 0, 0, 0), datetime(1960, 1, 2, 9, 0, 0)),
            ],
        ),
        (
            datetime(1960, 1, 1, 8, 0, 0),
            datetime(1960, 1, 3, 9, 0, 0),
            [
                (datetime(1960, 1, 1, 8, 0, 0), datetime(1960, 1, 2, 0, 0, 0)),
                (datetime(1960, 1, 2, 0, 0, 0), datetime(1960, 1, 3, 0, 0, 0)),
                (datetime(1960, 1, 3, 0, 0, 0), datetime(1960, 1, 3, 9, 0, 0)),
            ],
        ),
    ],
)
def test_gen_date_intervals(from_date, to_date, expected):
    dates = gen_date_intervals(from_date, to_date)
    assert list(dates) == expected


def midnights_apart(from_datetime, to_datetime) -> int:
    """Return the number of day changes between from_datetime and to_datetime."""
    return (to_datetime.date() - from_datetime.date()).days


@given(datetimes=from_to_datetime())
def test_date_tuples(datetimes):
    from_datetime, to_datetime = datetimes

    dates = list(gen_cut_dates(from_datetime, to_datetime))
    assert dates[0] == from_datetime
    assert dates[-1] == to_datetime

    num_days_apart = midnights_apart(from_datetime, to_datetime)
    # Remove from_datetime and to_datetime from count, remove 1 if to is midnight
    assert len(dates) - 2 == num_days_apart - (1 if is_midnight(to_datetime) else 0)

    # We always expect intervals to be exactly one day long
    for from_datetime, to_datetime in pairwise(dates[1:-1]):
        num_days_apart = midnights_apart(from_datetime, to_datetime)
        assert type(from_datetime) == datetime
        assert type(to_datetime) == datetime
        assert num_days_apart == 1
        assert (to_datetime - from_datetime).total_seconds() == 86400


@pytest.mark.parametrize(
    "date_time,expected",
    [
        (datetime(2022, 1, 1), "2022-01-01"),
        (datetime(100, 10, 1), "0100-10-01"),
        (datetime(10, 1, 10), "0010-01-10"),
        (datetime(1, 1, 1), "0001-01-01"),
    ],
)
def test_format_date_zero_fill(date_time: datetime, expected: str):
    assert format_date(date_time) == expected


@pytest.mark.parametrize(
    "date_time,expected",
    [
        (datetime(2022, 1, 1), "01.01.2022"),
        (datetime(2100, 10, 1), "01.10.2100"),
        (datetime(1000, 1, 10), "10.01.1000"),
        (datetime(3000, 10, 10), "10.10.3000"),
        (datetime(9999, 12, 31), "31.12.9999"),
    ],
)
def test_datetime_to_sd_date(date_time: datetime, expected: str):
    assert datetime_to_sd_date(date_time) == expected



import unittest
from typing import Any
from unittest import mock

from hypothesis import given
from hypothesis import strategies as st
from parameterized import parameterized
from sqlalchemy import Column
from sqlalchemy import create_engine
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import session
from sqlalchemy.orm import sessionmaker

from ..ad_exceptions import ImproperlyConfigured
from ..user_names import UserNameGen
from ..user_names import UserNameGenMethod2
from ..user_names import UserNameGenPermutation
from ..user_names import UserNameSet
from ..user_names import UserNameSetCSVFile
from ..user_names import UserNameSetCSVFileSubstring
from ..user_names import UserNameSetInAD
from ..user_names import UserNameSetInDatabase
from .mocks import MockADParameterReader
from .name_simulator import create_name


class TestUserNameGen(unittest.TestCase):
    _load_setting_path = "integrations.ad_integration.user_names.load_setting"

    @parameterized.expand(
        [
            ("UserNameGenMethod2", UserNameGenMethod2),
            ("UserNameGenUnknown", UserNameGenMethod2),
            ("UserNameGenPermutation", UserNameGenPermutation),
        ]
    )
    def test_override_implementation(self, name, expected_class):
        """Test that `get_implementation` returns the expected implementation
        configured by setting 'integrations.ad_writer.user_names.class'.
        """
        with mock.patch(self._load_setting_path, return_value=lambda: name):
            impl = UserNameGen.get_implementation()
            self.assertIsInstance(impl, expected_class)
            self.assertEqual(len(impl._loaded_occupied_name_sets), 0)

    def test_default_if_no_settings(self):
        """Test that `get_implementation` returns the default implementation
        if 'settings.json' could not be read.
        """

        def raise_filenotfound():
            raise FileNotFoundError()

        with mock.patch(self._load_setting_path, return_value=raise_filenotfound):
            impl = UserNameGen.get_implementation()
            self.assertIsInstance(impl, UserNameGenMethod2)
            self.assertEqual(len(impl._loaded_occupied_name_sets), 0)

    def test_load_occupied_names(self):
        """Test that we can configure `UserNameGen` to load one or more
        `UserNameSet` classes and add their individual sets of occupied names.
        """
        # Get `UserNameGen` implementation which loads occupied usernames from
        # `UserNameSet` (= an empty implementation.)
        impl = self._get_instance("UserNameSet")

        # Assert that we loaded two username sets (from AD and from the
        # `UserNameSet` specified in settings.)
        self.assertEqual(len(impl._loaded_occupied_name_sets), 2)

        # Assert that the first username set is the AD username set.
        self.assertIsInstance(impl._loaded_occupied_name_sets[0], UserNameSetInAD)

        # Assert that the second username set is the one we specified
        # via settings, and is an empty set.
        self.assertIsInstance(impl._loaded_occupied_name_sets[1], UserNameSet)
        self.assertSetEqual(set(impl._loaded_occupied_name_sets[1]), set())

        # Assert that the total set of occupied usernames is equal to
        # the usernames "found" by our mock `ADParameterReader`.
        self.assertSetEqual(
            impl.occupied_names,
            set(impl._loaded_occupied_name_sets[0]),
        )

    def test_load_occupied_names_invalid_name_raises(self):
        with self.assertRaises(ImproperlyConfigured):
            self._get_instance("InvalidUserNameSetName")

    def test_is_username_occupied_is_case_insensitive(self):
        impl = self._get_instance("UserNameSet")
        username = list(impl.occupied_names)[0]
        self.assertEqual(username, username.lower())
        self.assertTrue(impl.is_username_occupied(username.upper()))

    def _get_instance(self, usernameset_class_name: str) -> "UserNameGen":
        cls_names = [usernameset_class_name]
        settings = {
            f"{UserNameGen._setting_prefix}.extra_occupied_name_classes": cls_names
        }
        with self._patch_settings(settings):
            # Use mock `ADParameterReader` in `UserNameSetInAD`
            with mock.patch(
                "integrations.ad_integration.user_names.ADParameterReader",
                new=MockADParameterReader,
            ):
                impl = UserNameGen.get_implementation()
                impl.load_occupied_names()
                return impl

    def _patch_settings(self, settings: dict):
        return mock.patch(
            self._load_setting_path,
            new=lambda name, default=None: lambda: settings.get(name, default),
        )


# PIMJE is missing in specification documen
pmj_first = [
    "pmunj",
    "pmuje",
    "pjens",
    "pimuj",
    "pimje",
    "pijen",
    "piamj",
    "piaje",
    "piamu",
    "pimun",
    "pmunk",
    "jense",
    "mjens",
    "mujen",
    "munje",
    "munkj",
]
pmj_second = ["pmuj", "pimj", "pjen", "pije", "piaj", "pmje"]
pmj_third = ["pmj", "pje", "pij", "pim", "pmu"]
pmj_fourth = [
    "pmu2j",
    "pm2je",
    "p2jen",
    "p2muj",
    "p2mje",
    "pim2j",
    "pi2je",
    "pi2mj",
    "pia2j",
    "piam2",
    "pimu2",
    "pmun2",
]

pmj_fifth = [
    "pmj2",
    "pmu2",
    "pij2",
    "pim2",
    "pia2",
    "mun2",
    "jen2",
    "pm2j",
    "pi2j",
    "pi2m",
    "p2mj",
    "p2je",
    "p2mu",
]
pmj_sixth = [
    "pmunkj",
    "pimunj",
    "piamuj",
    "pmunje",
    "pmujen",
    "pmjens",
    "pjense",
    "pijens",
    "piajen",
]
pmj = [pmj_first, pmj_second, pmj_third, pmj_fourth, pmj_fifth, pmj_sixth]

akjpa_first = [
    "akjpa",
    "akjea",
    "akria",
    "akrja",
    "akrpa",
    "akpea",
    "ajepa",
    "ajpea",
    "apeta",
    "akran",
    "akjan",
    "akpan",
    "ajean",
    "ajpan",
    "apean",
    "aande",
    "ankpa",
    "ankja",
    "ankra",
    "anjpa",
    "anjea",
    "anpea",
    "ankan",
    "anjan",
    "anpan",
    "anand",
    "andka",
    "andja",
    "andpa",
    "andan",
    "andkr",
    "andkj",
    "andkp",
    "andje",
    "andjp",
    "andpe",
    "andea",
    "andek",
    "andej",
    "andep",
    "ankjp",
    "ankje",
    "ankrp",
    "ankrj",
    "ankri",
    "ankpe",
    "anjpe",
    "anjep",
    "anjen",
    "anpet",
    "akjpe",
    "akjep",
    "akrjp",
    "akrip",
    "akrij",
    "akrje",
    "akjen",
    "akrpe",
    "akpet",
    "ajpet",
    "ajepe",
    "ajenp",
    "ajens",
    "apete",
    "akris",
    "ander",
    "kande",
    "krand",
    "krian",
    "krisa",
    "kjand",
    "kjean",
    "kjena",
    "kjpan",
    "kjpea",
    "kpand",
    "kpean",
    "kpeta",
    "jande",
    "jeand",
    "jenan",
    "jensa",
    "jpand",
    "jpean",
    "jpeta",
    "pande",
    "peand",
    "petan",
    "petea",
]
akjpa_second = [
    "akra",
    "akja",
    "akpa",
    "ajea",
    "ajpa",
    "apea",
    "anka",
    "anja",
    "anpa",
    "aand",
    "anan",
    "anda",
    "akan",
    "ajan",
    "apan",
]
akjpa_third = [
    "aka",
    "aja",
    "apa",
    "aan",
    "ana",
    "ank",
    "anj",
    "anp",
    "akr",
    "akj",
    "akp",
    "aje",
    "ajp",
    "ape",
]
akjpa_fourth = [
    "akj2a",
    "akr2a",
    "akp2a",
    "aje2a",
    "ajp2a",
    "ape2a",
    "ak2an",
    "aj2an",
    "ap2an",
    "a2and",
    "a2kpa",
    "a2kja",
    "a2kra",
    "a2jpa",
    "a2jea",
    "a2pea",
    "a2kan",
    "a2jan",
    "a2pan",
    "ank2a",
    "anj2a",
    "anp2a",
    "an2an",
    "an2ka",
    "an2ja",
    "an2pa",
    "and2a",
    "ande2",
    "andk2",
    "andj2",
    "andp2",
    "ankr2",
    "ankj2",
    "ankp2",
    "anje2",
    "anjp2",
    "anpe2",
    "akjp2",
    "akri2",
    "akrj2",
    "akrp2",
    "akje2",
    "akpe2",
    "ajen2",
    "ajep2",
    "ajpe2",
    "apet2",
]
akjpa_fifth = [
    "aka2",
    "aja2",
    "apa2",
    "akj2",
    "akp2",
    "akr2",
    "ajp2",
    "aje2",
    "ape2",
    "ana2",
    "ank2",
    "anj2",
    "anp2",
    "and2",
    "kjp2",
    "kje2",
    "kri2",
    "jep2",
    "jpe2",
    "pet2",
    "ak2a",
    "aj2a",
    "ap2a",
    "an2a",
    "ak2j",
    "ak2p",
    "aj2p",
    "an2k",
    "an2j",
    "an2p",
    "a2ka",
    "a2ja",
    "a2pa",
    "a2an",
    "a2kj",
    "a2kp",
    "a2kr",
    "a2jp",
    "a2je",
    "a2pe",
]
akjpa_sixth = [
    "akrisa",
    "akrija",
    "akripa",
    "akrjea",
    "akrjpa",
    "akrpea",
    "akjena",
    "akjepa",
    "akjpea",
    "ajensa",
    "ajenpa",
    "ajepea",
    "ajpeta",
    "apetea",
    "ankria",
    "ankrja",
    "ankrpa",
    "ankjea",
    "ankjpa",
    "anjena",
    "anjepa",
    "anjpea",
    "anpeta",
    "andkra",
    "andkja",
    "andkpa",
    "andjea",
    "andjpa",
    "andpea",
    "andeka",
    "andeja",
    "andepa",
    "akrian",
    "akrjan",
    "akrpan",
    "akjean",
    "akjpan",
    "akpean",
    "ajenan",
    "ajepan",
    "ajpean",
    "apetan",
    "akrand",
    "akjand",
    "akpand",
    "ajeand",
    "ajpand",
    "apeand",
    "akande",
    "ajande",
    "apande",
    "aander",
    "anande",
    "andand",
    "andean",
    "andera",
]
akjpa = [
    akjpa_first,
    akjpa_second,
    akjpa_third,
    akjpa_fourth,
    akjpa_fifth,
    akjpa_sixth,
]

kj_first = ["kjens", "kajen", "karje", "karij", "jense"]
kj_second = ["kjen", "kaje", "karj"]
kj_third = ["kje", "kaj"]
kj_fourth = ["k2jen", "ka2je", "kar2j", "kari2"]
kj_fifth = ["kaj2", "kar2", "jen2", "ka2j", "k2je"]
kj_sixth = ["kjense", "kajens", "karjen", "karije", "karinj"]
kj = [kj_first, kj_second, kj_third, kj_fourth, kj_fifth, kj_sixth]

kmj_first = [
    "kmunj",
    "kmuje",
    "kjens",
    "kamuj",
    "kamje",
    "kajen",
    "karmj",
    "karje",
    "karmu",
    "karij",
    "karim",
    "kamun",
    "kmunk",
    "jense",
    "mjens",
    "mujen",
    "munje",
    "munkj",
]
kmj_second = ["kmuj", "kamj", "kjen", "kaje", "karj", "kmje"]
kmj_third = ["kmj", "kje", "kaj", "kam", "kmu"]
kmj_fourth = [
    "kmu2j",
    "km2je",
    "k2jen",
    "k2muj",
    "k2mje",
    "kam2j",
    "ka2je",
    "ka2mj",
    "kar2j",
    "kari2",
    "karm2",
    "kamu2",
    "kmun2",
]
kmj_fifth = [
    "kmj2",
    "kmu2",
    "kaj2",
    "kam2",
    "kar2",
    "mun2",
    "jen2",
    "km2j",
    "ka2j",
    "ka2m",
    "k2mj",
    "k2je",
    "k2mu",
]
kmj_sixth = [
    "kmunkj",
    "kamunj",
    "karmuj",
    "karimj",
    "kmunje",
    "kmujen",
    "kmjens",
    "kjense",
    "kajens",
    "karjen",
    "karije",
    "karinj",
]
kmj = [kmj_first, kmj_second, kmj_third, kmj_fourth, kmj_fifth, kmj_sixth]

ooha_first = [
    "oohoa",
    "oosta",
    "oosha",
    "oosar",
    "oohar",
    "ohoar",
    "oaroe",
    "oloha",
    "olosa",
    "olhoa",
    "oloar",
    "olhar",
    "olaro",
    "oleoa",
    "oleha",
    "olear",
    "oleos",
    "oleoh",
    "oleho",
    "oloho",
    "olosh",
    "olost",
    "olhos",
    "oosth",
    "oosho",
    "oohos",
    "ohost",
    "ooste",
    "osaro",
    "ostar",
    "ostea",
    "oharo",
    "ohosa",
    "haroe",
    "hoaro",
    "hosar",
    "hosta",
]
ooha_second = [
    "oosa",
    "ooha",
    "ohoa",
    "oloa",
    "olha",
    "oaro",
    "olar",
    "olea",
    "ooar",
    "ohar",
]
ooha_third = ["ooa", "oha", "oar", "ola", "olo", "olh", "oos", "ooh", "oho"]
ooha_fourth = [
    "ooh2a",
    "oos2a",
    "oho2a",
    "oo2ar",
    "oh2ar",
    "o2aro",
    "o2oha",
    "o2osa",
    "o2hoa",
    "o2oar",
    "o2har",
    "olo2a",
    "olh2a",
    "ol2ar",
    "ol2oa",
    "ol2ha",
    "ole2a",
    "oleo2",
    "oleh2",
    "olos2",
    "oloh2",
    "olho2",
    "oost2",
    "oosh2",
    "ooho2",
    "ohos2",
]
ooha_fifth = [
    "ooa2",
    "oha2",
    "ooh2",
    "oos2",
    "oho2",
    "ola2",
    "olo2",
    "olh2",
    "ole2",
    "ost2",
    "aro2",
    "oo2a",
    "oh2a",
    "ol2a",
    "oo2h",
    "ol2o",
    "ol2h",
    "o2oa",
    "o2ha",
    "o2ar",
    "o2oh",
    "o2os",
    "o2ho",
]
ooha_sixth = [
    "oostea",
    "oostha",
    "ooshoa",
    "oohosa",
    "ohosta",
    "olosta",
    "olosha",
    "olohoa",
    "olhosa",
    "oleosa",
    "oleoha",
    "olehoa",
    "oostar",
    "ooshar",
    "oohoar",
    "ohosar",
    "oosaro",
    "ooharo",
    "ohoaro",
    "ooaroe",
    "oharoe",
    "olaroe",
    "olearo",
]
ooha = [
    ooha_first,
    ooha_second,
    ooha_third,
    ooha_fourth,
    ooha_fifth,
    ooha_sixth,
]


class TestUserNameGenMethod2(unittest.TestCase):
    def _test_person(self, name, reference, max_level):
        name_creator = UserNameGenMethod2()
        success = True
        for level in range(0, max_level):
            for correct_user_name in reference[level]:
                user_name = name_creator.create_username(name)
                if user_name != correct_user_name:
                    success = False
                    print("Got: {}, expected: {}".format(user_name, correct_user_name))
        return success

    def test_pia_munk_jensen(self):
        name = ["Pia", "Munk", "Jensen"]
        for i in range(1, 7):
            test_result = self._test_person(name, pmj, i)
            self.assertTrue(test_result, "Priority: {}".format(i))

    def test_anders_kristian_jens_peter_andersen(self):
        name = ["Anders", "Kristian", "Jens", "Peter", "Andersen"]
        for i in range(1, 7):
            test_result = self._test_person(name, akjpa, i)
            self.assertTrue(test_result, "Priority: {}".format(i))

    def test_karina_jensen(self):
        name = ["Karina", "Jensen"]
        for i in range(1, 7):
            test_result = self._test_person(name, kj, i)
            self.assertTrue(test_result, "Priority: {}".format(i))

    def test_karina_munk_jensen(self):
        name = ["Karina", "Munk", "Jensen"]
        for i in range(1, 7):
            test_result = self._test_person(name, kmj, i)
            self.assertTrue(test_result, "Priority: {}".format(i))

    def test_Olê_Østergård_Høst_Ærøe(self):
        name = ["Olê", "Østergård", "Høst", "Ærøe"]
        for i in range(1, 7):
            test_result = self._test_person(name, ooha, i)
            self.assertTrue(test_result, "Priority: {}".format(i))

    def test_too_many_karina_jensens(self):
        """
        Tests behaviour of creating so many identical names, that
        the algorithm runs out of options. For 'Karina Jensen' this
        should happen after 89 attempts.
        """
        name = ["Karina", "Jensen"]
        name_creator = UserNameGenMethod2()
        for i in range(1, 88):
            name_creator.create_username(name)
        with self.assertRaisesRegex(RuntimeError, "Failed to create user name"):
            name_creator.create_username(name)

    def test_dry_run(self):
        """
        Tests that with dry_run set to True, we can keep adding the
        same person and never run into an answer of None.
        """
        name = ["Karina", "Jensen"]
        name_creator = UserNameGenMethod2()
        for i in range(1, 100):
            user_name = name_creator.create_username(name, dry_run=True)
        self.assertFalse(user_name is None)

    def test_multiple_names(self):
        """
        Test that we can create a large number of random names without hitting a
        bug that stops the program. Test succeeds if the code does not raise an
        exception.
        """
        name_creator = UserNameGenMethod2()
        for i in range(0, 250):
            name = create_name()
            name_creator.create_username(name)

    def test_name_fixer(self):
        """
        Test that the name fixer allows a-z and does not allow values outside this
        range.
        """
        name_creator = UserNameGenMethod2()
        name = ["Anders", "abzæ-{øå", "Andersen"]
        fixed_name = name_creator._name_fixer(name)
        expected_name = ["Anders", "abzaoa", "Andersen"]
        self.assertTrue(fixed_name == expected_name)

    def test_check_is_case_insensitive(self):
        name = ["Fornavn", "Efternavn"]
        name_creator = UserNameGenMethod2()
        # Generate username (no occupied names yet)
        first_username = name_creator.create_username(name)
        # Add upper-case version of generated username to list of occupied names
        name_creator.add_occupied_names({first_username.upper()})
        # Generate second username from same name
        second_username = name_creator.create_username(name)
        # Assert new username is different, even when case is ignored
        self.assertNotEqual(first_username.lower(), second_username.lower())


class TestUserNameGenPermutation(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.instance = UserNameGenPermutation()

    @given(
        st.lists(
            st.text(
                min_size=1,
                alphabet=st.characters(
                    whitelist_characters="bcdfghjklmnpqrstvwxz",
                    whitelist_categories=(),
                ),
            ),
            min_size=2,
        )
    )
    def test_valid_input(self, name):
        # If `name` has at least two items, each item being a string of at least
        # one consonant, we should be able to create a username.
        self.instance.create_username(name)

    def test_suffix_increments(self):
        name = ["B", "C", "D"]
        for expected_suffix in range(1, 100):
            username = self.instance.create_username(name)
            self.assertEqual(username, "bcd%d" % expected_suffix)

    def test_skips_names_already_taken(self):
        name = ["First Name", "Last-Name"]
        self.instance.add_occupied_names({"fnm1", "fnm4"})
        for expected_username in ("fnm2", "fnm3", "fnm5"):
            username = self.instance.create_username(name)
            self.assertEqual(username, expected_username)

    @parameterized.expand(
        [
            ("Abel Spendabel", "asp1"),  # First name starts with a vowel
            ("Erik Ejegod", "ejg1"),  # Both first name and last name start with a vowel
            ("Erik Episk Ejegod", "eps1"),  # All parts start with a vowel
            ("Gorm Den Gamle", "gdn1"),  # All parts start with a consonant
            ("Ba Ca Da", "bcd1"),  # All parts start with a consonant
            ("Theodor Fælgen", "tfl1"),  # Last name contains non-ASCII character
            ("Øjvind Ørn", "jrn1"),  # All parts begin with non-ASCII characters
            ("Ea Obe", "ebb1"),  # Last name contains just one consonant
            ("Ivan Aaaa", "ivn1"),  # Last name contains *only* vocals
            ("Ab Aaa", "abb1"),  # Only *one* consonant across *all* name parts
        ]
    )
    def test_by_example(self, name, expected_username):
        name = name.split(maxsplit=1)
        actual_username = self.instance.create_username(name)
        self.assertEqual(actual_username, expected_username)

    def test_check_is_case_insensitive(self):
        name = ["Fornavn", "Efternavn"]
        # Generate username (no occupied names yet)
        first_username = self.instance.create_username(name)
        # Add upper-case version of generated username to list of occupied names
        self.instance.add_occupied_names({first_username.upper()})
        # Generate second username from same name
        second_username = self.instance.create_username(name)
        # Assert new username is different, even when case is ignored
        self.assertNotEqual(first_username.lower(), second_username.lower())

    def test_max_iterations(self):
        with self.assertRaises(ValueError):
            self.instance.create_username(["A"])


class TestUserNameSet(unittest.TestCase):
    def test_contains_is_case_insensitive(self):
        username_set = UserNameSet()
        username_set._usernames = {"aaa", "BBB"}
        self.assertIn("Aaa", username_set)
        self.assertIn("bbb", username_set)


class _UserNameCSVFileHelperMixin:
    csv_path = "some/fs/path"
    csv_lines = [
        "%s,foo" % UserNameSetCSVFile._column_name,
        "abcd1234,",
        "efgh5678",
    ]

    def _get_instance(self, cls):
        load_setting_path = "integrations.ad_integration.user_names.load_setting"
        with mock.patch(load_setting_path, return_value=lambda: self.csv_path):
            with mock.patch("io.open") as mock_open:
                mock_open.return_value.__enter__.return_value = self.csv_lines
                instance = cls()
                instance._mock_open = mock_open
                return instance


class TestUserNameSetCSVFile(_UserNameCSVFileHelperMixin, unittest.TestCase):
    def test_can_read_csv(self):
        instance = self._get_instance(UserNameSetCSVFile)
        instance._mock_open.assert_called_once_with(
            self.csv_path,
            "r",
            encoding=UserNameSetCSVFile._encoding,
        )
        self.assertEqual(instance._usernames, {"abcd1234", "efgh5678"})

    def test_contains(self):
        instance = self._get_instance(UserNameSetCSVFile)
        self.assertIn("abcd1234", instance)
        self.assertNotIn("abc123", instance)

    def test_iter(self):
        instance = self._get_instance(UserNameSetCSVFile)
        self.assertSetEqual(instance._usernames, set(instance))


class TestUserNameCSVFileSubstring(_UserNameCSVFileHelperMixin, unittest.TestCase):
    def test_contains(self):
        instance = self._get_instance(UserNameSetCSVFileSubstring)
        # List of taken names contains an exact match for the new username
        self.assertIn("abcd1234", instance)
        # List of taken names contains a substring match for the new username
        self.assertIn("abcd12345678", instance)
        # New username is not a substring of any of the taken names
        self.assertNotIn("abcd", instance)


Base = declarative_base()  # type: Any


class MockedUserNameTable(Base):
    __tablename__ = "usernames"

    uid = Column(Integer, primary_key=True)
    username = Column(String)
    other_random_info = Column(String, default="test")


class TestUserNameSetInDatabase(unittest.TestCase):
    def setUp(self):
        self.engine = create_engine("sqlite://")
        self.session = sessionmaker(bind=self.engine, autoflush=False)()

        Base.metadata.create_all(self.engine)
        name1 = MockedUserNameTable(uid=1, username="Alice")
        self.session.add(name1)
        name2 = MockedUserNameTable(uid=2, username="Bob")
        self.session.add(name2)
        name3 = MockedUserNameTable(uid=3, username="Bob")
        self.session.add(name3)
        self.session.commit()

    def tearDown(self):
        Base.metadata.drop_all(self.engine)

    @mock.patch.object(
        UserNameSetInDatabase,
        "_get_settings",
        return_value=("", "usernames", "username"),
    )
    def test_load_from_database(self, settings_mock):
        with mock.patch.object(
            UserNameSetInDatabase, "_get_session", return_value=self.session
        ) as sql_mock:

            instance = UserNameSetInDatabase()
            assert sql_mock.call_count == 1
            assert instance._usernames == {"Alice", "Bob"}

    @mock.patch.object(UserNameSetInDatabase, "__init__", return_value=None)
    def test_get_session(self, init_mock):
        instance = UserNameSetInDatabase()
        s = instance._get_session("sqlite://")
        assert isinstance(s, session.Session)



import asyncio
import datetime
import logging
import os
import pickle
import time
import typing
from pathlib import Path

from dateutil.parser import parse as parse_date
from gql import gql
from more_itertools import first
from ra_utils.async_to_sync import async_to_sync
from ra_utils.job_settings import JobSettings
from raclients.graph.client import GraphQLClient
from raclients.graph.util import execute_paged
from tenacity import retry
from tenacity import stop_after_delay
from tenacity import wait_exponential

RETRY_MAX_TIME = 60 * 5


class GqlLoraCacheSettings(JobSettings):
    class Config:
        frozen = True

    use_new_cache: bool = False
    primary_manager_responsibility: str | None = None
    exporters_actual_state_manager_responsibility_class: str | None = None
    prometheus_pushgateway: str | None = "pushgateway"
    mox_base: str = "http://mo:5000/lora"
    std_page_size: int = 400

    def to_old_settings(self) -> dict[str, typing.Any]:
        """Convert our DatabaseSettings to a settings.json format.

        This serves to implement the adapter pattern, adapting from pydantic and its
        corresponding 12-factor configuration paradigm with environment variables, to
        the current functionality of the program, based on the settings format from
        settings.json.

        Eventually the entire settings-processing within the program should be
        rewritten with a process similar to what has been done for the SD integration,
        but it was out of scope for the change when this code was introduced.
        """

        settings = {
            "mora.base": self.mora_base,
            "mox.base": self.mox_base,
            "exporters": {
                "actual_state": {
                    "manager_responsibility_class": self.primary_manager_responsibility
                }
            },
            "use_new_cache": self.use_new_cache,
        }

        return settings


logger = logging.getLogger(__name__)


# used to correctly insert the object into the cache
def insert_obj(obj: dict, cache: dict) -> None:
    if obj is None:
        return
    if len(obj["obj"]) == 0:
        return
    if obj["uuid"] in cache:
        cache[obj["uuid"]].extend(obj["obj"])
    else:
        cache[obj["uuid"]] = obj["obj"]


# when getting a query using current, the object is a single dict. When getting a
# historic query it is a list of dicts. in order to uniformly process the two states
# we wrap the current object in a list
def align_current(item: dict) -> dict:
    item["obj"] = [item["obj"]]
    return item


# Does various transformations on a cache to align it with the old lora cache
def convert_dict(
    query_res: dict,
    resolve_object: bool = True,
    resolve_validity: bool = True,
    replace_dict: dict = {},
) -> dict:
    def replace(d: dict, dictionary: dict):
        for replace_from, replace_to in dictionary.items():
            d[replace_to] = d.pop(replace_from)
        return d

    def res_validity(d: dict):
        validity = d.pop("validity")
        if "from" in validity:
            if validity["from"]:
                d["from_date"] = str(parse_date(validity["from"]).date())
            else:
                d["from_date"] = str(datetime.datetime(1, 1, 1).date())
        if "to" in validity:
            if validity["to"]:
                d["to_date"] = str(parse_date(validity["to"]).date())
            else:
                d["to_date"] = "9999-12-31"
        return d

    uuid = query_res.pop("uuid")

    obj_list = []
    if resolve_object:
        for obj in query_res["obj"]:
            if obj is None:
                continue
            if resolve_validity:
                obj = res_validity(obj)
            obj = replace(obj, replace_dict)
            obj_list.append(obj)

        return {"uuid": uuid, "obj": obj_list}

    if resolve_validity:
        query_res = res_validity(query_res)

    return {uuid: replace(query_res, replace_dict)}


class GQLLoraCache:
    def __init__(
        self,
        resolve_dar: bool = True,
        full_history: bool = False,
        skip_past: bool = False,
        settings=None,
    ):
        msg = "Start LoRa cache, resolve dar: {}, full_history: {}"
        logger.info(msg.format(resolve_dar, full_history))
        if isinstance(settings, dict):
            settings = None
        self.resolve_dar = resolve_dar
        self.settings: GqlLoraCacheSettings = settings or GqlLoraCacheSettings()
        self.std_page_size = self.settings.std_page_size

        self.full_history = full_history
        self.skip_past = skip_past

        self.facets: dict = {}
        self.classes: dict = {}
        self.users: dict = {}
        self.units: dict = {}
        self.addresses: dict = {}
        self.engagements: dict = {}
        self.managers: dict = {}
        self.associations: dict = {}
        self.leaves: dict = {}
        self.roles: dict = {}
        self.itsystems: dict = {}
        self.it_connections: dict = {}
        self.kles: dict = {}
        self.related: dict = {}
        self.dar_cache: dict = {}

        self.gql_client_session: GraphQLClient

        self.org_uuid = self._get_org_uuid()

    def _setup_gql_client(self) -> GraphQLClient:
        return GraphQLClient(
            url=f"{self.settings.mora_base}/graphql/v3",
            client_id=self.settings.client_id,
            client_secret=self.settings.client_secret,
            auth_realm=self.settings.auth_realm,
            auth_server=self.settings.auth_server,
            httpx_client_kwargs={"timeout": 300},
            execute_timeout=300,
        )

    def get_historic_query(self) -> dict:
        params: typing.Dict[str, typing.Optional[str]] = {
            "to_date": str(datetime.datetime.now()),
            "from_date": str((datetime.datetime.now() - datetime.timedelta(minutes=1))),
        }
        if self.full_history:
            params["to_date"] = None
            params["from_date"] = None
        if self.skip_past:
            params["from_date"] = str(
                (datetime.datetime.now() - datetime.timedelta(minutes=1))
            )
        return params

    async def construct_query(
        self,
        query_type: str,
        simple_query: bool,
        query: str,
        variable_values: dict | None,
        page_size: int | None,
        offset: int,
    ):
        if variable_values is None:
            variable_values = {}

        query_footer = """
                                }
                            }
                        }"""
        query_header = ""

        if simple_query:
            query_header = (
                """
                                query ($limit: int, $offset: int) {
                                    page: """
                + query_type
                + """ (limit: $limit, offset: $offset){
                    """
            )
            query_footer = """
                                        }
                                    }"""

        else:
            if not self.full_history:
                query_header = (
                    """
                                query ($limit: int, $offset: int) {
                                    page: """
                    + query_type
                    + """ (limit: $limit, offset: $offset){
                        uuid
                        obj: current {"""
                )

            if self.full_history:
                query_header = (
                    """
                                query ($to_date: DateTime,
                                       $from_date: DateTime,
                                       $limit: int,
                                       $offset: int) {
                                    page: """
                    + query_type
                    + """ (from_date: $from_date,
                                                     to_date: $to_date,
                                                     limit: $limit,
                                                     offset: $offset) {
                        uuid
                        obj: objects {
                             """
                )
                variable_values.update(self.get_historic_query())

        return gql(query_header + query + query_footer), variable_values

    @retry(
        reraise=True,
        wait=wait_exponential(multiplier=1, min=4, max=10),
        stop=stop_after_delay(RETRY_MAX_TIME),
    )
    async def _execute_query(
        self,
        query: str,
        query_type: str,
        variable_values: typing.Optional[dict] = None,
        simple_query: bool = False,
        page_size: typing.Optional[int] = None,
        offset: int = 0,
    ):
        gql_query, gql_variable_values = await self.construct_query(
            query=query,
            query_type=query_type,
            variable_values=variable_values,
            page_size=page_size,
            offset=offset,
            simple_query=simple_query,
        )
        try:
            async for obj in execute_paged(
                gql_session=self.gql_client_session,
                document=gql_query,
                variable_values=gql_variable_values,
                per_page=(page_size or self.std_page_size),
            ):
                yield obj
        except Exception as e:
            logger.error(e)
            logger.error(offset)
            logger.error(gql_query)
            logger.error(gql_variable_values)
            raise e

    # Used to set a value in the __init__, if this was async, init would have to be
    # as well, which would mean that the new cache could only be initiated from
    # async functions
    def _get_org_uuid(self):
        query = gql(
            """
            query {
              org {
                uuid
              }
            }
            """
        )
        with GraphQLClient(
            url=f"{self.settings.mora_base}/graphql/v3",
            client_id=self.settings.client_id,
            client_secret=self.settings.client_secret,
            auth_realm=self.settings.auth_realm,
            auth_server=self.settings.auth_server,
            sync=True,
            httpx_client_kwargs={"timeout": None},
            execute_timeout=None,
        ) as session:
            org = session.execute(query)["org"]["uuid"]
        return org

    async def _cache_lora_facets(self) -> None:
        query = """
                uuid
                user_key
            """

        async for obj in self._execute_query(
            query=query,
            query_type="facets",
            simple_query=True,
        ):
            obj = convert_dict(obj, resolve_object=False, resolve_validity=False)
            self.facets.update(obj)

    async def _cache_lora_classes(self) -> None:
        query = """
                    uuid
                    user_key
                    name
                    scope
                    facet_uuid
            """

        dictionary = {"name": "title", "facet_uuid": "facet"}

        async for obj in self._execute_query(
            query=query,
            query_type="classes",
            simple_query=True,
        ):
            obj = convert_dict(
                obj,
                resolve_object=False,
                resolve_validity=False,
                replace_dict=dictionary,
            )
            self.classes.update(obj)

    async def _cache_lora_itsystems(self) -> None:
        query = """
                    uuid
                    user_key
                    name
            """

        async for obj in self._execute_query(
            query=query,
            query_type="itsystems",
            simple_query=True,
        ):
            obj = convert_dict(obj, resolve_object=False, resolve_validity=False)
            self.itsystems.update(obj)

    async def _cache_lora_users(self) -> None:
        query = """
                        uuid
                        cpr_no
                        user_key
                        name
                        givenname
                        surname
                        nickname
                        nickname_givenname
                        nickname_surname
                        validity {
                            from
                            to
                        }
            """

        dictionary = {
            "cpr_no": "cpr",
            "givenname": "fornavn",
            "surname": "efternavn",
            "name": "navn",
            "nickname": "kaldenavn",
            "nickname_givenname": "kaldenavn_fornavn",
            "nickname_surname": "kaldenavn_efternavn",
        }

        async for obj in self._execute_query(
            query=query,
            query_type="employees",
        ):
            if not self.full_history:
                obj = align_current(obj)

            obj = convert_dict(obj, replace_dict=dictionary)
            insert_obj(obj, self.users)

    async def _cache_lora_units(self) -> None:
        async def format_managers_and_location(qr: dict):
            def find_manager(managers: typing.List[dict]) -> str | None:
                if not managers:
                    return None
                if (
                    self.settings.exporters_actual_state_manager_responsibility_class
                    is None
                ):
                    return first(managers)["uuid"]
                return first(
                    map(
                        lambda m: m["uuid"],
                        filter(
                            lambda ma: (
                                self.settings.exporters_actual_state_manager_responsibility_class
                                in ma["responsibility_uuids"]
                            ),
                            managers,
                        ),
                    ),
                    None,
                )

            for man in qr["obj"]:
                if man is None:
                    continue
                if not self.full_history:
                    if man["manager_uuid"]:
                        man["manager_uuid"] = find_manager(man["manager_uuid"])
                    else:
                        man["manager_uuid"] = None

                    if man["acting_manager_uuid"]:
                        man["acting_manager_uuid"] = find_manager(
                            man["acting_manager_uuid"]
                        )
                    else:
                        man["acting_manager_uuid"] = None

                    ancestors = man.pop("ancestors")
                    location = man["name"]
                    for ancestor in ancestors:
                        location = ancestor["name"] + "\\" + location

                    man["location"] = location
            return qr

        if self.full_history:
            query = """
                            uuid
                            user_key
                            name
                            unit_type_uuid
                            org_unit_level_uuid
                            parent_uuid
                            org_unit_hierarchy_uuid: org_unit_hierarchy
                            validity {
                                from
                                to
                            }
                """
        else:

            query = """
                            uuid
                            user_key
                            name
                            unit_type_uuid
                            org_unit_level_uuid
                            parent_uuid
                            org_unit_hierarchy_uuid: org_unit_hierarchy
                            manager_uuid: managers(inherit: false) {
                                org_unit_uuid
                                responsibility_uuids
                                uuid
                            }
                            acting_manager_uuid: managers(inherit: true) {
                                org_unit_uuid
                                responsibility_uuids
                                uuid
                            }
                            ancestors {
                                name
                                uuid
                            }
                            validity {
                                from
                                to
                            }
                """

        dictionary = {
            "org_unit_level_uuid": "level",
            "org_unit_hierarchy_uuid": "org_unit_hierarchy",
            "parent_uuid": "parent",
            "unit_type_uuid": "unit_type",
        }

        async for obj in self._execute_query(
            query=query,
            query_type="org_units",
        ):
            if not self.full_history:
                obj = align_current(obj)

            for item in obj["obj"]:
                if item is None:
                    continue
                if item["parent_uuid"] == self.org_uuid:
                    item["parent_uuid"] = None

            obj = await format_managers_and_location(obj)

            obj = convert_dict(obj, replace_dict=dictionary)
            insert_obj(obj, self.units)

    async def _cache_lora_engagements(self) -> None:
        def collect_extensions(d: dict):
            for ext_obj in d["obj"]:
                if ext_obj is None:
                    continue
                ed = {}
                for i in range(1, 11):
                    ed[f"udvidelse_{i}"] = ext_obj.pop(f"extension_{i}")

                ext_obj["extensions"] = ed

            return d

        async def set_primary_boolean(res: dict) -> dict:
            for res_obj in res["obj"]:
                if res_obj is None:
                    continue
                prim = res_obj.pop("primary")
                res_obj["primary_boolean"] = False
                if prim and "scope" in prim:
                    res_obj["primary_boolean"] = int(prim.get("scope")) > 0
            return res

        query = """
                        uuid
                        employee_uuid
                        org_unit_uuid
                        fraction
                        user_key
                        engagement_type_uuid
                        primary_uuid
                        primary {
                            user_key
                            scope
                        }
                        job_function_uuid
                        extension_1
                        extension_2
                        extension_3
                        extension_4
                        extension_5
                        extension_6
                        extension_7
                        extension_8
                        extension_9
                        extension_10
                        validity {
                            from
                            to
                        }
            """

        dictionary = {
            "employee_uuid": "user",
            "engagement_type_uuid": "engagement_type",
            "job_function_uuid": "job_function",
            "org_unit_uuid": "unit",
            "primary_uuid": "primary_type",
        }

        async for obj in self._execute_query(
            query=query,
            query_type="engagements",
        ):
            if not self.full_history:
                obj = align_current(obj)

            obj = collect_extensions(obj)
            obj = await set_primary_boolean(obj)
            obj = convert_dict(obj, replace_dict=dictionary)
            insert_obj(obj, self.engagements)

    async def _cache_lora_roles(self) -> None:
        query = """
                        uuid
                        employee_uuid
                        org_unit_uuid
                        role_type_uuid
                        validity {
                            from
                            to
                        }
            """

        dictionary = {
            "employee_uuid": "user",
            "org_unit_uuid": "unit",
            "role_type_uuid": "role_type",
        }

        async for obj in self._execute_query(
            query=query,
            query_type="roles",
        ):
            if not self.full_history:
                obj = align_current(obj)

            obj = convert_dict(obj, replace_dict=dictionary)
            insert_obj(obj, self.roles)

    async def _cache_lora_leaves(self) -> None:
        query = """
                        uuid
                        employee_uuid
                        user_key
                        leave_type_uuid
                        engagement_uuid
                        validity  {
                            from
                            to
                        }
            """

        replace_dictionary = {
            "employee_uuid": "user",
            "leave_type_uuid": "leave_type",
            "engagement_uuid": "engagement",
        }

        async for obj in self._execute_query(
            query=query,
            query_type="leaves",
        ):
            if not self.full_history:
                obj = align_current(obj)

            obj = convert_dict(obj, replace_dict=replace_dictionary)
            insert_obj(obj, self.leaves)

    async def _cache_lora_it_connections(self) -> None:
        async def set_primary_boolean(res: dict) -> dict:
            for res_obj in res["obj"]:
                if res_obj is None:
                    continue
                prim = res_obj.pop("primary_uuid")
                if prim:
                    res_obj["primary_boolean"] = True
                else:
                    res_obj["primary_boolean"] = None

            return res

        query = """
                        uuid
                        employee_uuid
                        org_unit_uuid
                        user_key
                        itsystem_uuid
                        validity {
                            from
                            to
                        }
                        primary_uuid
            """
        dictionary = {
            "employee_uuid": "user",
            "itsystem_uuid": "itsystem",
            "org_unit_uuid": "unit",
            "user_key": "username",
        }

        async for obj in self._execute_query(
            query=query,
            query_type="itusers",
        ):
            if not self.full_history:
                obj = align_current(obj)

            obj = await set_primary_boolean(obj)
            obj = convert_dict(obj, replace_dict=dictionary)
            insert_obj(obj, self.it_connections)

    async def _cache_lora_kles(self) -> None:
        async def format_aspects(d: dict) -> dict:
            new_obj_list = []
            for kle_obj in d["obj"]:
                if kle_obj is None:
                    continue
                asp_list = []
                asps = kle_obj.pop("kle_aspect_uuids")
                for a_uuid in asps:
                    aspect = kle_obj.copy()
                    aspect["kle_aspect_uuid"] = a_uuid
                    asp_list.append(aspect)
                new_obj_list.extend(asp_list)
            d["obj"] = new_obj_list

            return d

        query = """
                        uuid
                        org_unit_uuid
                        kle_number_uuid
                        kle_aspect_uuids
                        user_key
                        validity {
                            from
                            to
                        }
            """

        dictionary = {
            "kle_aspect_uuid": "kle_aspect",
            "kle_number_uuid": "kle_number",
            "org_unit_uuid": "unit",
        }

        async for obj in self._execute_query(
            query=query,
            query_type="kles",
        ):
            if not self.full_history:
                obj = align_current(obj)

            obj = await format_aspects(obj)
            obj = convert_dict(obj, replace_dict=dictionary)
            insert_obj(obj, self.kles)

    async def _cache_lora_related(self) -> None:
        def format_related(d: dict):
            for rel_obj in d["obj"]:
                if rel_obj is None:
                    continue
                rel_uuids_list = rel_obj.pop("org_unit_uuids")
                for i in range(1, (len(rel_uuids_list) + 1)):
                    rel_obj[f"unit{i}_uuid"] = rel_uuids_list[i - 1]

            return d

        query = """
                        uuid
                        org_unit_uuids
                        validity {
                            from
                            to
                        }
            """

        async for obj in self._execute_query(
            query=query,
            query_type="related_units",
        ):
            if not self.full_history:
                obj = align_current(obj)

            obj = format_related(obj)

            obj = convert_dict(obj)

            insert_obj(obj, self.related)

    async def _cache_lora_managers(self) -> None:
        query = """
                        uuid
                        employee_uuid
                        org_unit_uuid
                        manager_type_uuid
                        manager_level_uuid
                        responsibility_uuids
                        validity {
                            from
                            to
                        }
            """

        dictionary = {
            "employee_uuid": "user",
            "manager_level_uuid": "manager_level",
            "manager_type_uuid": "manager_type",
            "responsibility_uuids": "manager_responsibility",
            "org_unit_uuid": "unit",
        }

        async for obj in self._execute_query(
            query=query,
            query_type="managers",
        ):
            if not self.full_history:
                obj = align_current(obj)

            obj = convert_dict(obj, replace_dict=dictionary)
            insert_obj(obj, self.managers)

    async def _cache_lora_associations(self) -> None:
        async def process_associations_helper(res: dict) -> dict:
            for res_obj in res["obj"]:
                if res_obj is None:
                    continue
                prim = res_obj.pop("primary")
                if prim:
                    if prim["user_key"] == "primary":
                        res_obj["primary_boolean"] = True
                    else:
                        res_obj["primary_boolean"] = False
                else:
                    res_obj["primary_boolean"] = None

                if not res_obj["it_user_uuid"]:
                    res_obj["job_function_uuid"] = None

            return res

        # TODO job_function_uuid  # Should be None if it_user_uuid is None,
        #  ask Mads why?! #48316 !764 5dc0d245 . Mads is always busy
        query = """
                        uuid
                        employee_uuid
                        org_unit_uuid
                        user_key
                        association_type_uuid
                        it_user_uuid
                        job_function_uuid
                        validity {
                            from
                            to
                        }
                        primary {
                            user_key
                        }
            """

        replace_dict = {
            "employee_uuid": "user",
            "association_type_uuid": "association_type",
            "it_user_uuid": "it_user",
            "job_function_uuid": "job_function",
            "org_unit_uuid": "unit",
        }

        async for obj in self._execute_query(
            query=query,
            query_type="associations",
        ):
            if not self.full_history:
                obj = align_current(obj)

            obj = await process_associations_helper(obj)
            obj = convert_dict(obj, replace_dict=replace_dict)
            insert_obj(obj, self.associations)

    async def _cache_lora_address(self):
        async def prep_address(d: dict) -> dict:
            for add_obj in d["obj"]:
                if add_obj is None:
                    continue
                scope = add_obj.pop("address_type")["scope"]

                add_obj["scope"] = scope_map[scope]

                if scope == "DAR":
                    add_obj["dar_uuid"] = add_obj["value"]
                    # We need to populate the dar cache, but do not want to
                    # resolve all addresses again, and the resolved dar address we do have is
                    # closest to betegnelse.
                    # We are willing to overwrite an address if it is already present, as it is the
                    # same address for each uuid
                    self.dar_cache[add_obj["value"]] = {"betegnelse": add_obj["name"]}
                else:
                    add_obj["dar_uuid"] = None

                if self.resolve_dar:
                    add_obj["value"] = add_obj.pop("name")
                else:
                    add_obj.pop("name")

            return d

        query = """
                        address_type_uuid
                        employee_uuid
                        org_unit_uuid
                        visibility_uuid
                        name
                        value
                        uuid
                        address_type {
                            scope
                        }
                        validity {
                            from
                            to
                        }
            """
        scope_map = {
            "EMAIL": "E-mail",
            "WWW": "Url",
            "PHONE": "Telefon",
            "PNUMBER": "P-nummer",
            "EAN": "EAN",
            "TEXT": "Text",
            "MULTIFIELD_TEXT": "Multifield_text",
            "DAR": "DAR",
        }

        replace_dict = {
            "employee_uuid": "user",
            "org_unit_uuid": "unit",
            "address_type_uuid": "adresse_type",
            "visibility_uuid": "visibility",
        }

        async for obj in self._execute_query(
            query=query,
            query_type="addresses",
        ):
            if not self.full_history:
                obj = align_current(obj)

            # Skip if both of the below are None, should really be an invariant in MO
            # employee_uuid
            # org_unit_uuid
            if any(
                map(
                    lambda o: o is not None
                    and o["employee_uuid"] is None
                    and o["org_unit_uuid"] is None,
                    obj["obj"],
                )
            ):
                continue

            obj = await prep_address(obj)
            obj = convert_dict(obj, replace_dict=replace_dict)
            insert_obj(obj, self.addresses)

    async def populate_cache_async(self, dry_run=None, skip_associations=False):
        """
        Perform the actual data import.
        :param skip_associations: If associations are not needed, they can be
        skipped for increased performance.
        :param dry_run: For testing purposes it is possible to read from cache.
        """
        if dry_run is None:
            dry_run = os.environ.get("USE_CACHED_LORACACHE", False)

        # Ensure that tmp/ exists
        Path("tmp/").mkdir(exist_ok=True)

        if self.full_history:
            if self.skip_past:
                facets_file = "tmp/facets_historic_skip_past.p"
                classes_file = "tmp/classes_historic_skip_past.p"
                users_file = "tmp/users_historic_skip_past.p"
                units_file = "tmp/units_historic_skip_past.p"
                addresses_file = "tmp/addresses_historic_skip_past.p"
                engagements_file = "tmp/engagements_historic_skip_past.p"
                managers_file = "tmp/managers_historic_skip_past.p"
                associations_file = "tmp/associations_historic_skip_past.p"
                leaves_file = "tmp/leaves_historic_skip_past.p"
                roles_file = "tmp/roles_historic_skip_past.p"
                itsystems_file = "tmp/itsystems_historic_skip_past.p"
                it_connections_file = "tmp/it_connections_historic_skip_past.p"
                kles_file = "tmp/kles_historic_skip_past.p"
                related_file = "tmp/related_historic_skip_past.p"
            else:
                facets_file = "tmp/facets_historic.p"
                classes_file = "tmp/classes_historic.p"
                users_file = "tmp/users_historic.p"
                units_file = "tmp/units_historic.p"
                addresses_file = "tmp/addresses_historic.p"
                engagements_file = "tmp/engagements_historic.p"
                managers_file = "tmp/managers_historic.p"
                associations_file = "tmp/associations_historic.p"
                leaves_file = "tmp/leaves_historic.p"
                roles_file = "tmp/roles_historic.p"
                itsystems_file = "tmp/itsystems_historic.p"
                it_connections_file = "tmp/it_connections_historic.p"
                kles_file = "tmp/kles_historic.p"
                related_file = "tmp/related_historic.p"
        else:
            facets_file = "tmp/facets.p"
            classes_file = "tmp/classes.p"
            users_file = "tmp/users.p"
            units_file = "tmp/units.p"
            addresses_file = "tmp/addresses.p"
            engagements_file = "tmp/engagements.p"
            managers_file = "tmp/managers.p"
            associations_file = "tmp/associations.p"
            leaves_file = "tmp/leaves.p"
            roles_file = "tmp/roles.p"
            itsystems_file = "tmp/itsystems.p"
            it_connections_file = "tmp/it_connections.p"
            kles_file = "tmp/kles.p"
            related_file = "tmp/related.p"

        if dry_run:
            with open(facets_file, "rb") as f:
                self.facets = pickle.load(f)
            with open(classes_file, "rb") as f:
                self.classes = pickle.load(f)
            with open(users_file, "rb") as f:
                self.users = pickle.load(f)
            with open(units_file, "rb") as f:
                self.units = pickle.load(f)
            with open(addresses_file, "rb") as f:
                self.addresses = pickle.load(f)
            with open(engagements_file, "rb") as f:
                self.engagements = pickle.load(f)
            with open(managers_file, "rb") as f:
                self.managers = pickle.load(f)

            if not skip_associations:
                with open(associations_file, "rb") as f:
                    self.associations = pickle.load(f)

            with open(leaves_file, "rb") as f:
                self.leaves = pickle.load(f)
            with open(roles_file, "rb") as f:
                self.roles = pickle.load(f)
            with open(itsystems_file, "rb") as f:
                self.itsystems = pickle.load(f)
            with open(it_connections_file, "rb") as f:
                self.it_connections = pickle.load(f)
            with open(kles_file, "rb") as f:
                self.kles = pickle.load(f)
            with open(related_file, "rb") as f:
                self.related = pickle.load(f)
            return

        t = time.time()  # noqa: F841
        msg = "Kørselstid: {:.1f}s, {} elementer, {:.0f}/s"  # noqa: F841
        async with self._setup_gql_client() as session:
            self.gql_client_session = session
            async with asyncio.TaskGroup() as tg:
                tg.create_task(self._cache_lora_address())
                tg.create_task(self._cache_lora_units())
                tg.create_task(self._cache_lora_engagements())
                tg.create_task(self._cache_lora_facets())
                tg.create_task(self._cache_lora_classes())
                tg.create_task(self._cache_lora_users())
                tg.create_task(self._cache_lora_managers())
                if not skip_associations:
                    tg.create_task(self._cache_lora_associations())
                tg.create_task(self._cache_lora_leaves())
                tg.create_task(self._cache_lora_roles())
                tg.create_task(self._cache_lora_itsystems())
                tg.create_task(self._cache_lora_it_connections())
                tg.create_task(self._cache_lora_kles())
                tg.create_task(self._cache_lora_related())

        def write_caches(cache, filename, name):
            logger.debug(f"writing {name}")
            if filename:
                with open(filename, "wb") as fw:
                    pickle.dump(cache, fw, pickle.DEFAULT_PROTOCOL)
            logger.debug(f"done with {name}")

        write_caches(self.facets, facets_file, "facets")
        write_caches(self.engagements, engagements_file, "engagements")
        write_caches(self.classes, classes_file, "classes")
        write_caches(self.users, users_file, "users")
        write_caches(self.units, units_file, "units")
        write_caches(self.managers, managers_file, "managers")
        write_caches(self.leaves, leaves_file, "leaves")
        write_caches(self.addresses, addresses_file, "addresses")
        write_caches(self.roles, roles_file, "roles")
        write_caches(self.itsystems, itsystems_file, "itsystems")
        write_caches(self.it_connections, it_connections_file, "it_connections")
        write_caches(self.kles, kles_file, "kles")
        write_caches(self.related, related_file, "related")

        if not skip_associations:
            write_caches(self.associations, associations_file, "associations")

    @async_to_sync
    async def populate_cache(self, dry_run=None, skip_associations=False):
        await self.populate_cache_async(
            dry_run=dry_run, skip_associations=skip_associations
        )

    def calculate_primary_engagements(self):
        # Needed for compatibility reasons
        pass

    def calculate_derived_unit_data(self):
        # Needed for compatibility reasons
        pass



# from collections import OrderedDict
from typing import Optional
from uuid import UUID

from integrations.opus import opus_helpers


def create_user(employee, org_uuid, uuid=None):
    if employee["firstName"] is None and employee["lastName"] is None:
        employee["firstName"] = "Ukendt"
        employee["lastName"] = "Ukendt"
        empl = dict(employee)
        empl.pop("cpr")
        print("No names on user")
        print(empl)
    cpr = opus_helpers.read_cpr(employee)
    payload = {
        "givenname": employee["firstName"],
        "surname": employee["lastName"],
        "cpr_no": cpr,
        "org": {"uuid": org_uuid},
    }
    if uuid is not None:
        payload["uuid"] = uuid
    return payload


def edit_engagement(data, mo_engagement_uuid):
    payload = {"type": "engagement", "uuid": mo_engagement_uuid, "data": data}
    return payload


def edit_role(validity, mo_role_uuid):
    data = {"validity": validity}
    payload = {"type": "role", "uuid": mo_role_uuid, "data": data}
    return payload


def create_engagement(
    employee, user_uuid, unit_uuid, job_function, engagement_type, primary, validity
):
    payload = {
        "type": "engagement",
        "org_unit": {"uuid": str(unit_uuid)},
        "person": {"uuid": user_uuid},
        "job_function": {"uuid": job_function},
        "primary": {"uuid": primary},
        "engagement_type": {"uuid": engagement_type},
        "user_key": employee["@id"],
        "validity": validity,
    }
    return payload


def create_role(employee, user_uuid, unit_uuid, role_type, validity):
    payload = {
        "type": "role",
        "org_unit": {"uuid": str(unit_uuid)},
        "person": {"uuid": user_uuid},
        "role_type": {"uuid": role_type},
        "user_key": "role_{}".format(employee["@id"]),
        "validity": validity,
    }
    return payload


def create_org_unit(unit, unit_user_key, unit_uuid, parent, unit_type, from_date):
    payload = {
        "uuid": unit_uuid,
        "user_key": unit_user_key,
        "name": unit["longName"],
        "parent": {"uuid": parent},
        "org_unit_type": {"uuid": unit_type},
        "validity": {"from": from_date, "to": None},
    }
    return payload


def edit_org_unit(unit, unit_user_key, unit_uuid, parent, unit_type, from_date):
    payload = {
        "type": "org_unit",
        "data": {
            "uuid": unit_uuid,
            "user_key": unit_user_key,
            "name": unit["longName"],
            "parent": {"uuid": parent},
            "org_unit_type": {"uuid": unit_type},
            "validity": {"from": from_date, "to": None},
        },
    }
    return payload


def terminate_detail(uuid, terminate_date, detail_type):
    payload = {"type": detail_type, "uuid": uuid, "validity": {"to": terminate_date}}
    return payload


def terminate_manager(uuid, terminate_date):
    payload = {"type": "manager", "uuid": uuid, "validity": {"to": terminate_date}}
    return payload


def connect_it_system_to_user(username, it_system, person_uuid, from_date):
    payload = {
        "type": "it",
        "user_key": username,
        "itsystem": {"uuid": it_system},
        "person": {"uuid": person_uuid},
        "validity": {"from": from_date, "to": None},
    }
    return payload


def edit_it_system_username(uuid, username, from_date):
    payload = {
        "type": "it",
        "uuid": uuid,
        "data": {
            "user_key": username,
            "validity": {"from": from_date, "to": None},
        },
    }
    return payload


def create_address(
    validity,
    address_type,
    value,
    unit_uuid=None,
    user_uuid=None,
    visibility: Optional[UUID] = None,
):
    if unit_uuid is None and user_uuid is None:
        raise Exception("Either unit or user must be specified")
    if unit_uuid and user_uuid:
        raise Exception("Only a unit or a person can be specified")

    payload = {
        "type": "address",
        "value": value,
        "address_type": address_type,
        "validity": validity,
        "visibility": visibility,
    }

    if unit_uuid is not None:
        payload["org_unit"] = {"uuid": unit_uuid}
    if user_uuid is not None:
        payload["person"] = {"uuid": user_uuid}
    return payload


def edit_address(data, mo_address_uuid):
    payload = {"type": "address", "uuid": mo_address_uuid, "data": data}
    return payload


def create_manager(
    user_key, unit, person, manager_type, level, responsibility, validity
):
    payload = {
        "type": "manager",
        "user_key": user_key,
        "org_unit": {"uuid": unit},
        "person": {"uuid": person},
        "manager_type": {"uuid": manager_type},
        "manager_level": {"uuid": level},
        "responsibility": [  # OPUS will never give more than one
            {"uuid": responsibility}
        ],
        "validity": validity,
    }
    return payload


def edit_manager(
    object_uuid, unit, person, manager_type, level, responsibility, validity
):
    payload = {
        "type": "manager",
        "uuid": object_uuid,
        "data": {
            "org_unit": {"uuid": unit},
            "person": {"uuid": person},
            "manager_type": {"uuid": manager_type},
            "manager_level": {"uuid": level},
            "responsibility": [  # OPUS will never give more than one
                {"uuid": responsibility}
            ],
            "validity": validity,
        },
    }
    return payload



# SPDX-FileCopyrightText: Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
"""Integration entrypoint."""
import logging

import sentry_sdk
from fastapi import APIRouter
from fastapi import FastAPI
from fastramqpi.config import Settings as FastRAMQPISettings
from fastramqpi.main import FastRAMQPI
from ra_utils.job_settings import JobSettings

from .trigger import trigger_router

logger = logging.getLogger(__name__)

fastapi_router = APIRouter()


class Settings(JobSettings):
    fastramqpi: FastRAMQPISettings

    class Config:
        frozen = True
        env_nested_delimiter = "__"


@fastapi_router.get("/")
async def index() -> dict[str, str]:
    return {"name": "sql_export"}


def create_fastramqpi(**kwargs) -> FastRAMQPI:
    settings: Settings = Settings(**kwargs)
    settings.start_logging_based_on_settings()
    if settings.sentry_dsn:
        sentry_sdk.init(dsn=settings.sentry_dsn)

    fastramqpi = FastRAMQPI(application_name="sql-export", settings=settings.fastramqpi)
    fastramqpi.get_app().include_router(trigger_router)

    app = fastramqpi.get_app()
    app.include_router(fastapi_router)

    return fastramqpi


def create_app(**kwargs) -> FastAPI:
    fastramqpi = create_fastramqpi(**kwargs)
    return fastramqpi.get_app()
